// Closure.h

#ifndef Closure_h
#define Closure_h

class ClosureBase
  {
	private:
		class Generic;
		typedef void (*GenericFunction)();
		typedef void (Generic::*GenericMethod)();

		void *object;
		GenericFunction function;
		GenericMethod method;
	
	public:
		template < class F, class M >
		ClosureBase( void *theObject, F theFunction, M theMethod )
		  : object( theObject ),
			 function( reinterpret_cast< GenericFunction >( theFunction ) ),
			 method( reinterpret_cast< GenericMethod >( theMethod ) )
		  {}

		template < class F, class M >
		ClosureBase( const void *theObject, F theFunction, M theMethod )
		  : object( const_cast<void *>( theObject ) ),
			 function( reinterpret_cast< GenericFunction >( theFunction ) ),
			 method( reinterpret_cast< GenericMethod >( theMethod ) )
		  {}
		
		static void *NoObject()								{ return 0; }
		static GenericFunction NoFunction()				{ return 0; }
		static GenericMethod NoMethod()					{ return 0; }
		
		template < class Obj > Obj& Object() const	{ return *static_cast<Obj*>( object ); }
		template < class F >   F    Function() const	{ return reinterpret_cast<F>( function ); }
		template < class M >   M    Method() const	{ return reinterpret_cast<const M>( method ); }
																	// the "const" pleases a CW 5.3 bug.
  };



template < class Signature > class Closure;




template < class R >
class Closure< R (*)() >: private ClosureBase
  {
	private:
		R (*caller)( const ClosureBase& );
		
		static R ReturnDefault( const ClosureBase& )
		  {
			return R();
		  }
		
		template <class F>
		static R CallFunction( const ClosureBase& base )
		  {
			F function = base.Function<F>();
			return function();
		  }
		
		template <class Obj, class F>
		static R CallFunctionOfObject( const ClosureBase& base )
		  {
			F function = base.Function<F>();
			Obj& object = base.Object<Obj>();
			return function( object );
		  }

		template <class Obj, class M>
		static R CallMethod( const ClosureBase& base )
			{
			 Obj& object = base.Object<Obj>();
			 M method = base.Method<M>();
			 return (object.*method)();
		   }
		
	public:
		Closure( const ClosureBase& theBase,
		         R (*theCaller)( const ClosureBase& ) )
		  : ClosureBase( theBase ),
			 caller( theCaller )
		  {}
		
		Closure()
		  : ClosureBase( NoObject(), NoFunction(), NoMethod() ),
		  	 caller( &ReturnDefault )
		  {}
		
		template< class FR >
		Closure( FR(*f)() )
		  : ClosureBase( NoObject(), f, NoMethod() ),
		  	 caller( &CallFunction< FR(*)() > )
		  {}
		
		template< class Obj, class FR, class FObj >
		Closure( Obj& obj, FR(*f)( FObj& ) )
		  : ClosureBase( &obj, f, NoMethod() ),
		  	 caller( &CallFunctionOfObject< Obj, FR(*)( FObj& ) > )
		  {}
		
		template< class Obj, class FR, class FObj >
		Closure( const Obj& obj, FR(*f)( const FObj& ) )
		  : ClosureBase( &obj, f, NoMethod() ),
		  	 caller( &CallFunctionOfObject< const Obj, FR(*)( const FObj& ) > )
		  {}
		
		template< class Obj, class MR, class MObj >
		Closure( Obj& obj, MR (MObj::*m)() )
		  : ClosureBase( &obj, NoFunction(), m ),
		  	 caller( &CallMethod< Obj, MR (MObj::*)() > )
		  {}
		
		template< class Obj, class MR, class MObj >
		Closure( const Obj& obj, MR (MObj::*m)() const )
		  : ClosureBase( &obj, NoFunction(), m ),
		  	 caller( &CallMethod< const Obj, MR (MObj::*)() const > )
		  {}
		
		R operator()() const
		  {
			return caller( *this );
		  }
  };

template< class FR >
Closure< FR(*)() >
MakeClosure( FR(*f)() )
  {
	return Closure< FR(*)() >( f );
  }

template< class Obj, class FR, class FObj >
Closure< FR(*)() >
MakeClosure( Obj& obj, FR(*f)( FObj& ) )
  {
	return Closure< FR(*)() >( obj, f );
  }

template< class Obj, class FR, class FObj >
Closure< FR(*)() >
MakeClosure( const Obj& obj, FR(*f)( const FObj& ) )
  {
	return Closure< FR(*)() >( obj, f );
  }

template< class Obj, class MR, class MObj >
Closure< MR(*)() >
MakeClosure( Obj& obj, MR (MObj::*m)() )
  {
	return Closure< MR(*)() >( obj, m );
  }

template< class Obj, class MR, class MObj >
Closure< MR(*)() >
MakeClosure( const Obj& obj, MR (MObj::*m)() const )
  {
	return Closure< MR(*)() >( obj, m );
  }




template < class R, class P0 >
class Closure< R (*)( P0 ) >: private ClosureBase
  {
	private:
		R (*caller)( const ClosureBase&, P0 );
		
		static R ReturnDefault( const ClosureBase&, P0 )
		  {
			return R();
		  }
		
		template <class F>
		static R CallFunction( const ClosureBase& base, P0 p0 )
		  {
			F function = base.Function<F>();
			return function( p0 );
		  }
		
		template <class Obj, class F>
		static R CallFunctionOfObject( const ClosureBase& base, P0 p0 )
		  {
			F function = base.Function<F>();
			Obj& object = base.Object<Obj>();
			return function( object, p0 );
		  }

		template <class Obj, class M>
		static R CallMethod( const ClosureBase& base, P0 p0 )
			{
			 Obj& object = base.Object<Obj>();
			 M method = base.Method<M>();
			 return (object.*method)( p0 );
		   }
		
	public:
		Closure( const ClosureBase& theBase,
		         R (*theCaller)( const ClosureBase&, P0 ) )
		  : ClosureBase( theBase ),
			 caller( theCaller )
		  {}
		
		Closure()
		  : ClosureBase( NoObject(), NoFunction(), NoMethod() ),
		  	 caller( &ReturnDefault )
		  {}
		
		template< class FR, class FP0 >
		Closure( FR(*f)( FP0 ) )
		  : ClosureBase( NoObject(), f, NoMethod() ),
		  	 caller( &CallFunction< FR(*)( FP0 ) > )
		  {}
		
		template< class Obj, class FR, class FObj, class FP0 >
		Closure( Obj& obj, FR(*f)( FObj&, FP0 ) )
		  : ClosureBase( &obj, f, NoMethod() ),
		  	 caller( &CallFunctionOfObject< Obj, FR(*)( FObj&, FP0 ) > )
		  {}
		
		template< class Obj, class FR, class FObj, class FP0 >
		Closure( const Obj& obj, FR(*f)( const FObj&, FP0 ) )
		  : ClosureBase( &obj, f, NoMethod() ),
		  	 caller( &CallFunctionOfObject< const Obj, FR(*)( const FObj&, FP0 ) > )
		  {}
		
		template< class Obj, class MR, class MObj, class MP0 >
		Closure( Obj& obj, MR (MObj::*m)( MP0 ) )
		  : ClosureBase( &obj, NoFunction(), m ),
		  	 caller( &CallMethod< Obj, MR (MObj::*)( MP0 ) > )
		  {}
		
		template< class Obj, class MR, class MObj, class MP0 >
		Closure( const Obj& obj, MR (MObj::*m)( MP0 ) const )
		  : ClosureBase( &obj, NoFunction(), m ),
		  	 caller( &CallMethod< const Obj, MR (MObj::*)( MP0 ) const > )
		  {}
		
		R operator()( P0 p0 ) const
		  {
			return caller( *this, p0 );
		  }
  };

template< class FR, class FP0 >
Closure< FR(*)( FP0 ) >
MakeClosure( FR(*f)( FP0 ) )
  {
	return Closure< FR(*)( FP0 ) >( f );
  }

template< class Obj, class FR, class FObj, class FP0 >
Closure< FR(*)( FP0 ) >
MakeClosure( Obj& obj, FR(*f)( FObj&, FP0 ) )
  {
	return Closure< FR(*)( FP0 ) >( obj, f );
  }

template< class Obj, class FR, class FObj, class FP0 >
Closure< FR(*)( FP0 ) >
MakeClosure( const Obj& obj, FR(*f)( const FObj&, FP0 ) )
  {
	return Closure< FR(*)( FP0 ) >( obj, f );
  }

template< class Obj, class MR, class MObj, class MP0 >
Closure< MR(*)( MP0 ) >
MakeClosure( Obj& obj, MR (MObj::*m)( MP0 ) )
  {
	return Closure< MR(*)( MP0 ) >( obj, m );
  }

template< class Obj, class MR, class MObj, class MP0 >
Closure< MR(*)( MP0 ) >
MakeClosure( const Obj& obj, MR (MObj::*m)( MP0 ) const )
  {
	return Closure< MR(*)( MP0 ) >( obj, m );
  }




template < class R, class P0, class P1 >
class Closure< R (*)( P0, P1 ) >: private ClosureBase
  {
	private:
		R (*caller)( const ClosureBase&, P0, P1 );
		
		static R ReturnDefault( const ClosureBase&, P0, P1 )
		  {
			return R();
		  }
		
		template <class F>
		static R CallFunction( const ClosureBase& base, P0 p0, P1 p1 )
		  {
			F function = base.Function<F>();
			return function( p0, p1 );
		  }
		
		template <class Obj, class F>
		static R CallFunctionOfObject( const ClosureBase& base, P0 p0, P1 p1 )
		  {
			F function = base.Function<F>();
			Obj& object = base.Object<Obj>();
			return function( object, p0, p1 );
		  }

		template <class Obj, class M>
		static R CallMethod( const ClosureBase& base, P0 p0, P1 p1 )
			{
			 Obj& object = base.Object<Obj>();
			 M method = base.Method<M>();
			 return (object.*method)( p0, p1 );
		   }
		
	public:
		Closure( const ClosureBase& theBase,
		         R (*theCaller)( const ClosureBase&, P0, P1 ) )
		  : ClosureBase( theBase ),
			 caller( theCaller )
		  {}
		
		Closure()
		  : ClosureBase( NoObject(), NoFunction(), NoMethod() ),
		  	 caller( &ReturnDefault )
		  {}
		
		template< class FR, class FP0, class FP1 >
		Closure( FR(*f)( FP0, FP1 ) )
		  : ClosureBase( NoObject(), f, NoMethod() ),
		  	 caller( &CallFunction< FR(*)( FP0, FP1 ) > )
		  {}
		
		template< class Obj, class FR, class FObj, class FP0, class FP1 >
		Closure( Obj& obj, FR(*f)( FObj&, FP0, FP1 ) )
		  : ClosureBase( &obj, f, NoMethod() ),
		  	 caller( &CallFunctionOfObject< Obj, FR(*)( FObj&, FP0, FP1 ) > )
		  {}
		
		template< class Obj, class FR, class FObj, class FP0, class FP1 >
		Closure( const Obj& obj, FR(*f)( const FObj&, FP0, FP1 ) )
		  : ClosureBase( &obj, f, NoMethod() ),
		  	 caller( &CallFunctionOfObject< const Obj, FR(*)( const FObj&, FP0, FP1 ) > )
		  {}
		
		template< class Obj, class MR, class MObj, class MP0, class MP1 >
		Closure( Obj& obj, MR (MObj::*m)( MP0, MP1 ) )
		  : ClosureBase( &obj, NoFunction(), m ),
		  	 caller( &CallMethod< Obj, MR (MObj::*)( MP0, MP1 ) > )
		  {}
		
		template< class Obj, class MR, class MObj, class MP0, class MP1 >
		Closure( const Obj& obj, MR (MObj::*m)( MP0, MP1 ) const )
		  : ClosureBase( &obj, NoFunction(), m ),
		  	 caller( &CallMethod< const Obj, MR (MObj::*)( MP0, MP1 ) const > )
		  {}
		
		R operator()( P0 p0, P1 p1 ) const
		  {
			return caller( *this, p0, p1 );
		  }
  };

template< class FR, class FP0, class FP1 >
Closure< FR(*)( FP0, FP1 ) >
MakeClosure( FR(*f)( FP0, FP1 ) )
  {
	return Closure< FR(*)( FP0, FP1 ) >( f );
  }

template< class Obj, class FR, class FObj, class FP0, class FP1 >
Closure< FR(*)( FP0, FP1 ) >
MakeClosure( Obj& obj, FR(*f)( FObj&, FP0, FP1 ) )
  {
	return Closure< FR(*)( FP0, FP1 ) >( obj, f );
  }

template< class Obj, class FR, class FObj, class FP0, class FP1 >
Closure< FR(*)( FP0, FP1 ) >
MakeClosure( const Obj& obj, FR(*f)( const FObj&, FP0, FP1 ) )
  {
	return Closure< FR(*)( FP0, FP1 ) >( obj, f );
  }

template< class Obj, class MR, class MObj, class MP0, class MP1 >
Closure< MR(*)( MP0, MP1 ) >
MakeClosure( Obj& obj, MR (MObj::*m)( MP0, MP1 ) )
  {
	return Closure< MR(*)( MP0, MP1 ) >( obj, m );
  }

template< class Obj, class MR, class MObj, class MP0, class MP1 >
Closure< MR(*)( MP0, MP1 ) >
MakeClosure( const Obj& obj, MR (MObj::*m)( MP0, MP1 ) const )
  {
	return Closure< MR(*)( MP0, MP1 ) >( obj, m );
  }




template < class R, class P0, class P1, class P2 >
class Closure< R (*)( P0, P1, P2 ) >: private ClosureBase
  {
	private:
		R (*caller)( const ClosureBase&, P0, P1, P2 );
		
		static R ReturnDefault( const ClosureBase&, P0, P1, P2 )
		  {
			return R();
		  }
		
		template <class F>
		static R CallFunction( const ClosureBase& base, P0 p0, P1 p1, P2 p2 )
		  {
			F function = base.Function<F>();
			return function( p0, p1, p2 );
		  }
		
		template <class Obj, class F>
		static R CallFunctionOfObject( const ClosureBase& base, P0 p0, P1 p1, P2 p2 )
		  {
			F function = base.Function<F>();
			Obj& object = base.Object<Obj>();
			return function( object, p0, p1, p2 );
		  }

		template <class Obj, class M>
		static R CallMethod( const ClosureBase& base, P0 p0, P1 p1, P2 p2 )
			{
			 Obj& object = base.Object<Obj>();
			 M method = base.Method<M>();
			 return (object.*method)( p0, p1, p2 );
		   }
		
	public:
		Closure( const ClosureBase& theBase,
		         R (*theCaller)( const ClosureBase&, P0, P1, P2 ) )
		  : ClosureBase( theBase ),
			 caller( theCaller )
		  {}
		
		Closure()
		  : ClosureBase( NoObject(), NoFunction(), NoMethod() ),
		  	 caller( &ReturnDefault )
		  {}
		
		template< class FR, class FP0, class FP1, class FP2 >
		Closure( FR(*f)( FP0, FP1, FP2 ) )
		  : ClosureBase( NoObject(), f, NoMethod() ),
		  	 caller( &CallFunction< FR(*)( FP0, FP1, FP2 ) > )
		  {}
		
		template< class Obj, class FR, class FObj, class FP0, class FP1, class FP2 >
		Closure( Obj& obj, FR(*f)( FObj&, FP0, FP1, FP2 ) )
		  : ClosureBase( &obj, f, NoMethod() ),
		  	 caller( &CallFunctionOfObject< Obj, FR(*)( FObj&, FP0, FP1, FP2 ) > )
		  {}
		
		template< class Obj, class FR, class FObj, class FP0, class FP1, class FP2 >
		Closure( const Obj& obj, FR(*f)( const FObj&, FP0, FP1, FP2 ) )
		  : ClosureBase( &obj, f, NoMethod() ),
		  	 caller( &CallFunctionOfObject< const Obj, FR(*)( const FObj&, FP0, FP1, FP2 ) > )
		  {}
		
		template< class Obj, class MR, class MObj, class MP0, class MP1, class MP2 >
		Closure( Obj& obj, MR (MObj::*m)( MP0, MP1, MP2 ) )
		  : ClosureBase( &obj, NoFunction(), m ),
		  	 caller( &CallMethod< Obj, MR (MObj::*)( MP0, MP1, MP2 ) > )
		  {}
		
		template< class Obj, class MR, class MObj, class MP0, class MP1, class MP2 >
		Closure( const Obj& obj, MR (MObj::*m)( MP0, MP1, MP2 ) const )
		  : ClosureBase( &obj, NoFunction(), m ),
		  	 caller( &CallMethod< const Obj, MR (MObj::*)( MP0, MP1, MP2 ) const > )
		  {}
		
		R operator()( P0 p0, P1 p1, P2 p2 ) const
		  {
			return caller( *this, p0, p1, p2 );
		  }
  };

template< class FR, class FP0, class FP1, class FP2 >
Closure< FR(*)( FP0, FP1, FP2 ) >
MakeClosure( FR(*f)( FP0, FP1, FP2 ) )
  {
	return Closure< FR(*)( FP0, FP1, FP2 ) >( f );
  }

template< class Obj, class FR, class FObj, class FP0, class FP1, class FP2 >
Closure< FR(*)( FP0, FP1, FP2 ) >
MakeClosure( Obj& obj, FR(*f)( FObj&, FP0, FP1, FP2 ) )
  {
	return Closure< FR(*)( FP0, FP1, FP2 ) >( obj, f );
  }

template< class Obj, class FR, class FObj, class FP0, class FP1, class FP2 >
Closure< FR(*)( FP0, FP1, FP2 ) >
MakeClosure( const Obj& obj, FR(*f)( const FObj&, FP0, FP1, FP2 ) )
  {
	return Closure< FR(*)( FP0, FP1, FP2 ) >( obj, f );
  }

template< class Obj, class MR, class MObj, class MP0, class MP1, class MP2 >
Closure< MR(*)( MP0, MP1, MP2 ) >
MakeClosure( Obj& obj, MR (MObj::*m)( MP0, MP1, MP2 ) )
  {
	return Closure< MR(*)( MP0, MP1, MP2 ) >( obj, m );
  }

template< class Obj, class MR, class MObj, class MP0, class MP1, class MP2 >
Closure< MR(*)( MP0, MP1, MP2 ) >
MakeClosure( const Obj& obj, MR (MObj::*m)( MP0, MP1, MP2 ) const )
  {
	return Closure< MR(*)( MP0, MP1, MP2 ) >( obj, m );
  }




template < class R, class P0, class P1, class P2, class P3 >
class Closure< R (*)( P0, P1, P2, P3 ) >: private ClosureBase
  {
	private:
		R (*caller)( const ClosureBase&, P0, P1, P2, P3 );
		
		static R ReturnDefault( const ClosureBase&, P0, P1, P2, P3 )
		  {
			return R();
		  }
		
		template <class F>
		static R CallFunction( const ClosureBase& base, P0 p0, P1 p1, P2 p2, P3 p3 )
		  {
			F function = base.Function<F>();
			return function( p0, p1, p2, p3 );
		  }
		
		template <class Obj, class F>
		static R CallFunctionOfObject( const ClosureBase& base, P0 p0, P1 p1, P2 p2, P3 p3 )
		  {
			F function = base.Function<F>();
			Obj& object = base.Object<Obj>();
			return function( object, p0, p1, p2, p3 );
		  }

		template <class Obj, class M>
		static R CallMethod( const ClosureBase& base, P0 p0, P1 p1, P2 p2, P3 p3 )
			{
			 Obj& object = base.Object<Obj>();
			 M method = base.Method<M>();
			 return (object.*method)( p0, p1, p2, p3 );
		   }
		
	public:
		Closure( const ClosureBase& theBase,
		         R (*theCaller)( const ClosureBase&, P0, P1, P2, P3 ) )
		  : ClosureBase( theBase ),
			 caller( theCaller )
		  {}
		
		Closure()
		  : ClosureBase( NoObject(), NoFunction(), NoMethod() ),
		  	 caller( &ReturnDefault )
		  {}
		
		template< class FR, class FP0, class FP1, class FP2, class FP3 >
		Closure( FR(*f)( FP0, FP1, FP2, FP3 ) )
		  : ClosureBase( NoObject(), f, NoMethod() ),
		  	 caller( &CallFunction< FR(*)( FP0, FP1, FP2, FP3 ) > )
		  {}
		
		template< class Obj, class FR, class FObj, class FP0, class FP1, class FP2, class FP3 >
		Closure( Obj& obj, FR(*f)( FObj&, FP0, FP1, FP2, FP3 ) )
		  : ClosureBase( &obj, f, NoMethod() ),
		  	 caller( &CallFunctionOfObject< Obj, FR(*)( FObj&, FP0, FP1, FP2, FP3 ) > )
		  {}
		
		template< class Obj, class FR, class FObj, class FP0, class FP1, class FP2, class FP3 >
		Closure( const Obj& obj, FR(*f)( const FObj&, FP0, FP1, FP2, FP3 ) )
		  : ClosureBase( &obj, f, NoMethod() ),
		  	 caller( &CallFunctionOfObject< const Obj, FR(*)( const FObj&, FP0, FP1, FP2, FP3 ) > )
		  {}
		
		template< class Obj, class MR, class MObj, class MP0, class MP1, class MP2, class MP3 >
		Closure( Obj& obj, MR (MObj::*m)( MP0, MP1, MP2, MP3 ) )
		  : ClosureBase( &obj, NoFunction(), m ),
		  	 caller( &CallMethod< Obj, MR (MObj::*)( MP0, MP1, MP2, MP3 ) > )
		  {}
		
		template< class Obj, class MR, class MObj, class MP0, class MP1, class MP2, class MP3 >
		Closure( const Obj& obj, MR (MObj::*m)( MP0, MP1, MP2, MP3 ) const )
		  : ClosureBase( &obj, NoFunction(), m ),
		  	 caller( &CallMethod< const Obj, MR (MObj::*)( MP0, MP1, MP2, MP3 ) const > )
		  {}
		
		R operator()( P0 p0, P1 p1, P2 p2, P3 p3 ) const
		  {
			return caller( *this, p0, p1, p2, p3 );
		  }
  };

template< class FR, class FP0, class FP1, class FP2, class FP3 >
Closure< FR(*)( FP0, FP1, FP2, FP3 ) >
MakeClosure( FR(*f)( FP0, FP1, FP2, FP3 ) )
  {
	return Closure< FR(*)( FP0, FP1, FP2, FP3 ) >( f );
  }

template< class Obj, class FR, class FObj, class FP0, class FP1, class FP2, class FP3 >
Closure< FR(*)( FP0, FP1, FP2, FP3 ) >
MakeClosure( Obj& obj, FR(*f)( FObj&, FP0, FP1, FP2, FP3 ) )
  {
	return Closure< FR(*)( FP0, FP1, FP2, FP3 ) >( obj, f );
  }

template< class Obj, class FR, class FObj, class FP0, class FP1, class FP2, class FP3 >
Closure< FR(*)( FP0, FP1, FP2, FP3 ) >
MakeClosure( const Obj& obj, FR(*f)( const FObj&, FP0, FP1, FP2, FP3 ) )
  {
	return Closure< FR(*)( FP0, FP1, FP2, FP3 ) >( obj, f );
  }

template< class Obj, class MR, class MObj, class MP0, class MP1, class MP2, class MP3 >
Closure< MR(*)( MP0, MP1, MP2, MP3 ) >
MakeClosure( Obj& obj, MR (MObj::*m)( MP0, MP1, MP2, MP3 ) )
  {
	return Closure< MR(*)( MP0, MP1, MP2, MP3 ) >( obj, m );
  }

template< class Obj, class MR, class MObj, class MP0, class MP1, class MP2, class MP3 >
Closure< MR(*)( MP0, MP1, MP2, MP3 ) >
MakeClosure( const Obj& obj, MR (MObj::*m)( MP0, MP1, MP2, MP3 ) const )
  {
	return Closure< MR(*)( MP0, MP1, MP2, MP3 ) >( obj, m );
  }

#endif
