// Array.h

#ifndef Array_h
#define Array_h

#include <algorithm>
#include <stdexcept>
#ifndef ArrayIterator_h
#include "ArrayIterator.h"
#endif
#ifndef ArrayRef_h
#include "ArrayRef.h"
#endif

template < class Element, std::size_t length > class ArrayBase;

template < class Element >
class ArrayBase< const Element, 0 >
  {
	public:
		ArrayRef< const Element > Reference() const
		  { 
			return ArrayRef< const Element >( reinterpret_cast<const Element *>( this ),
														 reinterpret_cast<const Element *>( this ) );
		  }
  };

template < class Element >
class ArrayBase< Element, 0 >
  {
	public:
		ArrayRef< Element > Reference()
		  { 
			return ArrayRef< Element >( reinterpret_cast<Element *>( this ),
												 reinterpret_cast<Element *>( this ) );
		  }

		ArrayRef< const Element > Reference() const
		  { 
			return ArrayRef< const Element >( reinterpret_cast<const Element *>( this ),
														 reinterpret_cast<const Element *>( this ) );
		  }
  };

template < class Element, std::size_t length >
class ArrayBase< const Element, length >
  {
	public:															// To allow aggregate initialization; avoid it otherwise
		Element array[ length ];
	
	public:
		ArrayRef< const Element > Reference() const		{ return ArrayRef< const Element >( array, array + length ); }
  };

template < class Element, std::size_t length >
class ArrayBase
  {
	public:															// To allow aggregate initialization; avoid it otherwise
		Element array[ length ];
	
	public:
		ArrayRef< Element > Reference()						{ return ArrayRef< Element >( array, array + length ); }
		ArrayRef< const Element > Reference() const		{ return ArrayRef< const Element >( array, array + length ); }
  };

template < class Element, std::size_t length >
class Array: private ArrayBase< Element, length >
  {
	private:
		typedef ArrayBase< Element, length > Inherited;
		
	public:
		typedef ArrayRef< Element > 							ArrayReference;
		typedef ArrayRef< const Element > 					ConstArrayReference;

		typedef ArrayReference::value_type					value_type;
		typedef ArrayReference::pointer						pointer;
		typedef ArrayReference::const_pointer				const_pointer;
		typedef ArrayReference::reference					reference;
		typedef ArrayReference::const_reference			const_reference;
		typedef ArrayReference::difference_type			difference_type;
		typedef ArrayReference::size_type					size_type;
		typedef ArrayReference::const_iterator				const_iterator;
		typedef ArrayReference::iterator						iterator;
		typedef ArrayReference::reverse_iterator			reverse_iterator;
		typedef ArrayReference::const_reverse_iterator	const_reverse_iterator;

		Array()																{}
		Array( const ConstArrayReference& s )						{ Reference() = s; }
		// Plus the implicitly defined copy constructor
		
		Array& operator=( const ConstArrayReference& s )		{ Reference() = s; return *this; }
		// Plus the implicitly defined copy assignment

		operator ArrayReference()										{ return Reference(); }
		operator ConstArrayReference() const						{ return Reference(); }
		
		iterator begin()													{ return Reference().begin(); }
		const_iterator begin() const									{ return Reference().begin(); }
		
		iterator end()														{ return Reference().end(); }
		const_iterator end() const										{ return Reference().end(); }
		
		iterator IteratorTo( Element *p )							{ return Reference().IteratorTo(p); }
		const_iterator IteratorTo( const Element *p ) const	{ return Reference().IteratorTo(p); }

		void swap( const ArrayReference& b )						{ Reference().swap(b); }
		
		static size_type size()											{ return length; }
		static size_type max_size()									{ return length; }
		static bool empty()												{ return length == 0; }

		reference front()													{ return Reference().front(); }
		const_reference front() const									{ return Reference().front(); }

		reference back()													{ return Reference().back(); }
		const_reference back() const									{ return Reference().back(); }

		reference operator[]( size_type n )							{ return Reference()[n]; }
		const_reference operator[]( size_type n ) const			{ return Reference()[n]; }

		reference at( size_type n )									{ return Reference().at(n); }
		const_reference at( size_type n ) const					{ return Reference().at(n); }

		bool operator==( const ConstArrayReference& b ) const	{ return Reference() == b; }
		bool operator< ( const ConstArrayReference& b ) const	{ return Reference() >  b; }
		bool operator> ( const ConstArrayReference& b ) const	{ return Reference() < b; }
		bool operator!=( const ConstArrayReference& b ) const	{ return Reference() != b; }
		bool operator<=( const ConstArrayReference& b ) const	{ return Reference() <= b; }
		bool operator>=( const ConstArrayReference& b ) const	{ return Reference() >= b; }

		reverse_iterator rbegin()										{ return Reference().rbegin(); }
		const_reverse_iterator rbegin() const						{ return Reference().rbegin(); }
		
		reverse_iterator rend()											{ return Reference().rend(); }
		const_reverse_iterator rend() const							{ return Reference().rend(); }
  };

namespace std
  {
	template < class Element, std::size_t length >
	inline void swap( Array< Element, length >& a,
						   Array< Element, length >& b )
	  {
		a.swap( b );
	  }
  }

#endif
