// ArrayIterator.h

#ifndef ArrayIterator_h
#define ArrayIterator_h

#include <cstddef>
#include <iterator>
#ifndef ConfigureIteratorSafety_h
#include "ConfigureIteratorSafety.h"
#endif
#ifndef Assert_h
#include "Assert.h"
#endif

template < class Element, bool isSafe > class ArrayIteratorSafety;
template < class Element > class ArrayIterator;

template < class Element >
class ArrayIteratorSafety< Element, true >
  {
	private:
		const Element *begin;
		const Element *end;
	
	public:
		ArrayIteratorSafety()																: begin( 0 ), end( 0 )				{}
		explicit ArrayIteratorSafety( const Element *b, const Element *e )	: begin( b ), end( e )				{}
		
		bool CanPointTo( const Element *p ) const										{ return begin <= p && p <= end; }
		
		bool CanIncrement( const Element *p ) const									{ return p < end; }
		bool CanDecrement( const Element *p ) const									{ return p > begin; }
		
		bool CanAdd( const Element *p, std::ptrdiff_t n ) const					{ return begin - p <= n && n <= end - p; }
		bool CanSubtract( const Element *p, std::ptrdiff_t n ) const			{ return p - end <= n && n <= p - begin; }

		bool ComparableTo( const Element *p ) const									{ return CanPointTo( p ); }
		
		bool CanDereference( const Element *p ) const								{ return p != end; }
  };

template < class Element >
class ArrayIteratorSafety< Element, false >
  {
	public:
		ArrayIteratorSafety()																{}
		explicit ArrayIteratorSafety( const Element *, const Element * )		{}
		bool CanPointTo( const Element * ) const										{ return true; }
		bool CanIncrement( const Element * ) const									{ return true; }
		bool CanDecrement( const Element * ) const									{ return true; }
		bool CanAdd( const Element *, std::ptrdiff_t ) const						{ return true; }
		bool CanSubtract( const Element *, std::ptrdiff_t ) const				{ return true; }
		bool ComparableTo( const Element * ) const									{ return true; }
		bool CanDereference( const Element * ) const									{ return true; }
  };

template < class Element >
class ArrayIterator< const Element >: private ArrayIteratorSafety< Element, useSafeIterators >
  {
	public:
		typedef Element															value_type;
		typedef std::ptrdiff_t													difference_type;
		typedef const Element *													pointer;
		typedef const Element&													reference;
		typedef std::random_access_iterator_tag							iterator_category;
		
		typedef ArrayIterator< const Element >								const_iterator;
		typedef ArrayIteratorSafety< Element, useSafeIterators > 	Safety;

	private:
		pointer p;
	
	public:
		ArrayIterator()														: p( 0 )										{ Assert( CanPointTo( p ) ); }
		ArrayIterator( pointer position, const Safety& safety )	: Safety( safety ), p( position )	{ Assert( CanPointTo( p ) ); }
		
		operator pointer() const											{ return p; }
		
		reference operator*() const										{ Assert( CanDereference(p) ); return *p; }
		pointer operator->() const											{ Assert( CanDereference(p) ); return p; }
		
		const_iterator& operator++()										{ Assert( CanIncrement(p) );   ++p; return *this; }
		const_iterator& operator--()										{ Assert( CanDecrement(p) );   --p; return *this; }

		const_iterator operator++(int)									{ Assert( CanIncrement(p) );   return const_iterator( p++, *this ); }				
		const_iterator operator--(int)									{ Assert( CanDecrement(p) );   return const_iterator( p--, *this ); }
		
		const_iterator& operator+=( difference_type n )				{ Assert( CanAdd(p,n) );       p += n; return *this; }
		const_iterator& operator-=( difference_type n )				{ Assert( CanSubtract(p,n) );  p -= n; return *this; }

		const_iterator operator+( difference_type n ) const		{ Assert( CanAdd(p,n) );       return const_iterator( p + n, *this ); }
		const_iterator operator-( difference_type n ) const		{ Assert( CanSubtract(p,n) );  return const_iterator( p - n, *this ); }
		
		difference_type operator-( const_iterator& b ) const		{ Assert( ComparableTo(b.p) ); return p - b.p; }
		
		bool operator==( const const_iterator& b ) const			{ Assert( ComparableTo(b.p) ); return p == b.p; }
		bool operator< ( const const_iterator& b ) const			{ Assert( ComparableTo(b.p) ); return p <  b.p; }

		// Boilerplate comparisons
			bool operator> ( const const_iterator& b ) const		{ return b < *this; }
			bool operator!=( const const_iterator& b ) const		{ return !( *this == b ); }
			bool operator<=( const const_iterator& b ) const		{ return !( *this > b ); }
			bool operator>=( const const_iterator& b ) const		{ return !( *this < b ); }
		
		// Boilerplate for random access
			reference operator[]( difference_type n ) const			{ return *operator+(n); }
  };

template < class Element >
class ArrayIterator: private ArrayIteratorSafety< Element, useSafeIterators >
  {
	public:
		typedef Element															value_type;
		typedef std::ptrdiff_t													difference_type;
		typedef Element *															pointer;
		typedef Element&															reference;
		typedef std::random_access_iterator_tag							iterator_category;
		
		typedef ArrayIterator< const Element >								const_iterator;
		typedef ArrayIterator< Element >										iterator;
		typedef ArrayIteratorSafety< Element, useSafeIterators >		Safety;

	private:
		pointer p;
	
	public:
		ArrayIterator()														: p( 0 )										{ Assert( CanPointTo( p ) ); }
		ArrayIterator( pointer position, const Safety& safety )	: Safety( safety ), p( position )	{ Assert( CanPointTo( p ) ); }
		
		operator pointer() const											{ return p; }
		operator const_iterator() const									{ return const_iterator( p, *this ); }

		reference operator*() const										{ Assert( CanDereference(p) ); return *p; }
		pointer operator->() const											{ Assert( CanDereference(p) ); return p; }
		
		iterator& operator++()												{ Assert( CanIncrement(p) );   ++p; return *this; }
		iterator& operator--()												{ Assert( CanDecrement(p) );   --p; return *this; }

		iterator operator++(int)											{ Assert( CanIncrement(p) );   return iterator( p++, *this ); }				
		iterator operator--(int)											{ Assert( CanDecrement(p) );   return iterator( p--, *this ); }
		
		iterator& operator+=( difference_type n )						{ Assert( CanAdd(p,n) );       p += n; return *this; }
		iterator& operator-=( difference_type n )						{ Assert( CanSubtract(p,n) );  p -= n; return *this; }

		iterator operator+( difference_type n ) const				{ Assert( CanAdd(p,n) );       return iterator( p + n, *this ); }
		iterator operator-( difference_type n ) const				{ Assert( CanSubtract(p,n) );  return iterator( p - n, *this ); }

		// Boilerplate non-const iterator comparisons
			difference_type operator-( const_iterator& b ) const	{ return operator const_iterator() - b; }
			
			bool operator==( const const_iterator& b ) const		{ return operator const_iterator() == b;   }
			bool operator!=( const const_iterator& b ) const		{ return operator const_iterator() != b; }
			bool operator< ( const const_iterator& b ) const		{ return operator const_iterator() < b; }
			bool operator> ( const const_iterator& b ) const		{ return operator const_iterator() > b; }
			bool operator<=( const const_iterator& b ) const		{ return operator const_iterator() <= b; }
			bool operator>=( const const_iterator& b ) const		{ return operator const_iterator() >= b; }
		
		// Boilerplate for random access
			reference operator[]( difference_type n ) const			{ return *operator+(n); }
  };

template < class Element >
ArrayIterator<Element> operator+( typename ArrayIterator<Element>::difference_type n,
											 const ArrayIterator<Element>& a )
  {
	return a + n;
  }

#endif
