// ArrayRef.h

#ifndef ArrayRef_h
#define ArrayRef_h

#include <cstddef>
#include <stdexcept>
#include <algorithm>
#ifndef Assert_h
#include "Assert.h"
#endif
#ifndef ArrayIterator_h
#include "ArrayIterator.h"
#endif

template < class Element > class ArrayPtr;
template < class Element > class ArrayRef;

template < class Element >
ArrayPtr< Element > MakeArrayPtr( Element *begin, Element *end );

template < class Element >
class ArrayRef< const Element >
  {
	friend class ArrayPtr< const Element >;
	
	public:
		typedef Element									value_type;
		typedef const Element *							pointer;
		typedef const Element *							const_pointer;
		typedef const Element&							reference;
		typedef const Element&							const_reference;
		typedef std::ptrdiff_t							difference_type;
		typedef std::size_t								size_type;

		typedef ArrayIterator< const value_type >	const_iterator;
		typedef ArrayIterator< const value_type >	iterator;
		
	private:
		pointer b;
		pointer e;

		iterator::Safety Safety() const										{ return iterator::Safety( b, e ); }
		
		ArrayRef()																	: b( 0 ), e( 0 )			{}
		bool Null() const															{ return b == 0; }
		void Reseat( const ArrayRef& s )										{ b = s.b; e = s.e; }
		const ArrayRef *Address() const										{ return this; }	// since & is overloaded
		
		// not implemented:
			ArrayRef& operator=( const ArrayRef& );
		
	public:
		ArrayRef( pointer begin, pointer end )								: b( begin ), e( end )	{ Assert( !Null() ); }
		// Plus the implicitly-generated copy constructor
		
		//These aren't supported by CW Pro 5
		//template < size_type n > ArrayRef( const Element(&a)[n] )	: b( a ), e( a+n )		{}
		//template < size_type n > ArrayRef( Element(&a)[n] )			: b( a ), e( a+n )		{}
		
		iterator begin() const													{ return iterator( b, Safety() ); }		
		iterator end() const														{ return iterator( e, Safety() ); }
		iterator IteratorTo( pointer p ) const								{ return iterator( p, Safety() ); }

		size_type size() const													{ return e - b; }
		size_type max_size() const												{ return e - b; }
		bool empty() const														{ return e == b; }

		ArrayPtr< const Element > operator&() const						{ return MakeArrayPtr( b, e ); }
		
		// Boilerplate optional sequence operations (partial)
			reference front() const												{ return *begin(); }
			reference back() const												{ return *--end(); }
			reference operator[]( size_type n ) const						{ return *( begin() + n ); }

			reference at( size_type n ) const
			  {
				if ( n >= size() )
					throw std::out_of_range( "Array index out of range" );
				
				return operator[]( n );
			  }

		// Boilerplate comparisons
			bool operator==( const ArrayRef& b ) const
			  {
				return size() == b.size() && std::equal( begin(), end(), b.begin() );
			  }

			bool operator<( const ArrayRef& b ) const
			  {
				return std::lexicographical_compare( begin(),   end(),
																 b.begin(), b.end() );
			  }

			bool operator> ( const ArrayRef& b ) const					{ return b < *this; }
			bool operator!=( const ArrayRef& b ) const					{ return !( *this == b ); }
			bool operator<=( const ArrayRef& b ) const					{ return !( *this > b ); }
			bool operator>=( const ArrayRef& b ) const					{ return !( *this < b ); }

		// Boilerplate reversibility
			typedef std::reverse_iterator< iterator > reverse_iterator;
			typedef std::reverse_iterator< const_iterator > const_reverse_iterator;

			reverse_iterator rbegin() const									{ return reverse_iterator( end() ); }
			reverse_iterator rend() const										{ return reverse_iterator( begin() ); }
  };

template < class Element >
class ArrayRef
  {
	friend class ArrayPtr< Element >;
	
	public:
		typedef Element									value_type;
		typedef Element *									pointer;
		typedef const Element *							const_pointer;
		typedef Element&									reference;
		typedef const Element&							const_reference;
		typedef std::ptrdiff_t							difference_type;
		typedef std::size_t								size_type;

		typedef ArrayIterator< const value_type >	const_iterator;
		typedef ArrayIterator< value_type >			iterator;
		
	private:
		pointer b;
		pointer e;

		iterator::Safety Safety() const										{ return iterator::Safety( b, e ); }
		
		ArrayRef()																	: b( 0 ), e( 0 )			{}
		bool Null() const															{ return b == 0; }
		void Reseat( const ArrayRef& s )										{ b = s.b; e = s.e; }
		const ArrayRef *Address() const										{ return this; }	// since & is overloaded
		
	public:
		ArrayRef( pointer begin, pointer end )								: b( begin ), e( end )	{ Assert( !Null() ); }
		// Plus the implicitly-generated copy constructor

		//This isn't supported by CW Pro 5
		//template < size_type n > ArrayRef( Element(&a)[n] )			: b( a ), e( a+n )		{}
		
		const ArrayRef& operator=( const ArrayRef& s ) const;
		void swap( const ArrayRef& s ) const;
		
		operator ArrayRef< const value_type >() const					{ return ArrayRef< const value_type >( b, e ); }
		
		iterator begin() const													{ return iterator( b, Safety() ); }		
		iterator end() const														{ return iterator( e, Safety() ); }
		iterator IteratorTo( pointer p ) const								{ return iterator( p, Safety() ); }

		size_type size() const													{ return e - b; }
		size_type max_size() const												{ return e - b; }
		bool empty() const														{ return e == b; }

		ArrayPtr< const Element > operator&() const						{ return MakeArrayPtr( b, e ); }

		// Boilerplate optional sequence operations (partial)
			reference front() const												{ return *begin(); }
			reference back() const												{ return *--end(); }
			reference operator[]( size_type n ) const						{ return *( begin() + n ); }

			reference at( size_type n ) const
			  {
				if ( n >= size() )
					throw std::out_of_range( "Array index out of range" );
				
				return operator[]( n );
			  }

		// Boilerplate comparisons
			bool operator==( const ArrayRef& b ) const
			  {
				return size() == b.size() && std::equal( begin(), end(), b.begin() );
			  }

			bool operator<( const ArrayRef& b ) const
			  {
				return std::lexicographical_compare( begin(),   end(),
																 b.begin(), b.end() );
			  }

			bool operator> ( const ArrayRef& b ) const					{ return b < *this; }
			bool operator!=( const ArrayRef& b ) const					{ return !( *this == b ); }
			bool operator<=( const ArrayRef& b ) const					{ return !( *this > b ); }
			bool operator>=( const ArrayRef& b ) const					{ return !( *this < b ); }

		// Boilerplate reversibility
			typedef std::reverse_iterator< iterator > reverse_iterator;
			typedef std::reverse_iterator< const_iterator > const_reverse_iterator;

			reverse_iterator rbegin() const									{ return reverse_iterator( end() ); }
			reverse_iterator rend() const										{ return reverse_iterator( begin() ); }
  };

template < class Element >
const ArrayRef< Element >& ArrayRef< Element >::operator=( const ArrayRef& b ) const
  {
	Assert( size() == b.size() );
	
	iterator left = begin();
	iterator right = b.begin();

	for ( ; left != end() && right != b.end(); ++left,++right )
		*left = *right;
	
	return *this;
  }

template < class Element >
void ArrayRef< Element >::swap( const ArrayRef& b ) const
  {
	Assert( size() == b.size() );
	
	iterator left = begin();
	iterator right = b.begin();

	for ( ; left != end() && right != b.end(); ++left,++right )
		std::swap( *left, *right );
  }

namespace std  // Overloading in namespace std isn't strictly conforming,
  { 				// but may be a good idea anyway.
	template < class Element >
	inline void swap( ArrayRef< Element >& a, ArrayRef< Element >& b )
	  {
		a.swap( b );
	  }
  }

#endif
