// LinkedList.h

#ifndef LinkedList_h
#define LinkedList_h

#include <algorithm>
#include <functional>
#ifndef ConfigureIteratorSafety_h
#include "ConfigureIteratorSafety.h"
#endif
#ifndef LinkBody_h
#include "LinkBody.h"
#endif
#ifndef LinkedListBase_h
#include "LinkedListBase.h"
#endif

template < class Element >
class LinkedList: private LinkedListBase< useSafeIterators >
  {
	private:
		typedef LinkedListBase< useSafeIterators > Base;
		typedef LinkBody< Element >					 Body;
	
	public:
		class Position;
		class Link;
		class const_iterator;
		class iterator;

		typedef Base::size_type				size_type;
		typedef Base::difference_type		difference_type;
		typedef Body::value_type			value_type;
		typedef Body::reference				reference;
		typedef Body::const_reference		const_reference;
		
		class Position: private Base::Link
		  {
			friend class Link;
			friend class const_iterator;
			friend class iterator;
			friend class LinkedList;
			
			private:
				Position()															{}
				Position( Position& s )											: Base::Link( s )			{}
			
			public:
				Position& operator=( Position& s )							{ Base::Link::operator=( s ); return *this; }
				void swap( Position& s )										{ Base::Link::swap( s ); }
				
				void Unlink()														{ Base::Link::Unlink(); }
				
				bool operator==( const Position& b ) const				{ return this == &b; }
				bool operator!=( const Position& b ) const				{ return this != &b; }
		  };
		
		class Link: private Body,
						private Position
		  {
			friend class const_iterator;
			friend class iterator;
			friend class LinkedList;
				
			public:
				typedef Body::value_type			value_type;
				typedef Body::reference				reference;
				typedef Body::const_reference		const_reference;
				typedef Body::pointer				pointer;
				typedef Body::const_pointer		const_pointer;
				typedef LinkedList::Position		PositionType;
				
				Link()																{}
				Link( reference s )												: Body( s )								{}
				Link( const_reference s )										: Body( s )								{}
				Link( Link& s )													: Body( s ), PositionType( s )	{}
				Link( ResourceTransfer<Link> s )								: Body( *s ), PositionType( *s )	{}

				operator ResourceTransfer<Link>()							{ return *this; }

				Link& operator=( Link& s )
				  {
					Body::operator=( s );
					Position::operator=( s );
					return *this;
				  }
				
				Link& operator=( ResourceTransfer<Link> s )				{ return operator=( *s ); }

				void swap( Link& s )												{ Body::swap( s ); PositionType::swap( s ); }
				
				PositionType& Position()										{ return *this; }
				const PositionType& Position() const						{ return *this; }
				
				void Unlink()														{ PositionType::Unlink(); }
				
				reference operator*()											{ return Body::operator*(); }
				const_reference operator*() const							{ return Body::operator*(); }
				pointer operator->()												{ return Body::operator->(); }
				const_pointer operator->() const								{ return Body::operator->(); }
				
				// Should be, but CW can't handle:
					// using Body::operator*;
					// using Body::operator->;
		  };
		
		class const_iterator: private Base::const_iterator
		  {
			friend class iterator;
			friend class LinkedList;
			
			private:
				typedef Base::const_iterator Inherited;
				const_iterator( const Inherited& s )						: Inherited( s )	{}
			
			public:
				typedef Inherited::difference_type		difference_type;
				typedef Body::value_type					value_type;
				typedef Body::const_pointer				pointer;
				typedef Body::const_reference				reference;
				typedef Inherited::iterator_category	iterator_category;
				typedef LinkedList::Link 					LinkType;
				typedef LinkedList::Position 				PositionType;
				
				const_iterator()													: p( 0 )				{}
				
				const LinkType& Link() const									{ return static_cast< const LinkType& >( Inherited::operator*() ); }
				const PositionType& Position() const						{ return Link(); }
				
				reference operator*() const									{ return Link().Body::operator*(); }
				pointer operator->() const										{ return Link().Body::operator->(); }
				
				const_iterator& operator++()									{ Inherited::operator++(); return *this; }
				const_iterator& operator--()									{ Inherited::operator--(); return *this; }

				const_iterator operator++(int)								{ return Inherited::operator++(0); }				
				const_iterator operator--(int)								{ return Inherited::operator--(0); }
				
				bool operator==( const const_iterator& b ) const		{ return Inherited::operator==( b ); }
				bool operator!=( const const_iterator& b ) const		{ return Inherited::operator!=( b ); }
		  };
		
		class iterator: private Base::iterator
		  {
			friend class LinkedList;
			
			private:
				typedef Base::iterator Inherited;
				iterator( const Inherited& s )								: Inherited( s )	{}

			public:
				typedef Inherited::difference_type		difference_type;
				typedef Body::value_type					value_type;
				typedef Body::pointer						pointer;
				typedef Body::reference						reference;
				typedef Inherited::iterator_category	iterator_category;
				typedef LinkedList::Link 					LinkType;
				typedef LinkedList::Position 				PositionType;

				iterator()															: p( 0 )				{}
				
				operator const_iterator() const								{ return const_iterator( p ); }
				
				LinkType& Link() const											{ return static_cast< LinkType& >( Inherited::operator*() ); }
				PositionType& Position() const								{ return Link(); }
				
				reference operator*() const									{ return Link().Body::operator*(); }
				pointer operator->() const										{ return Link().Body::operator->(); }
				
				iterator& operator++()											{ Inherited::operator++(); return *this; }
				iterator& operator--()											{ Inherited::operator--(); return *this; }

				iterator operator++(int)										{ return Inherited::operator++(0); }				
				iterator operator--(int)										{ return Inherited::operator--(0); }
				
				bool operator==( const const_iterator& b ) const		{ return Inherited::operator==( b ); }
				bool operator!=( const const_iterator& b ) const		{ return Inherited::operator!=( b ); }
		  };
	
	public:
		LinkedList()															   {}
		LinkedList( LinkedList& source )										: Base(  source )	{}
		LinkedList( ResourceTransfer<LinkedList> source )				: Base( *source )	{}
		LinkedList( LinkedList& s, iterator what )						: Base( s, what ) {}
		LinkedList( LinkedList& s, iterator b, iterator e )			: Base( s, b, e ) {}
		
		template < class InputIterator >
		LinkedList( InputIterator b, InputIterator e )					{ while ( b != e ) push_back( *b++ ); }
		
		operator ResourceTransfer<LinkedList>()							{ return *this; }
	
		LinkedList& operator=( LinkedList& s )								{ Base::operator=( s ); return *this; }
		LinkedList& operator=( ResourceTransfer<LinkedList> s )		{ return operator=( *s ); }

		void swap( LinkedList& s )												{ Base::swap( s ); }

		iterator begin()															{ return Base::begin(); }
		const_iterator begin() const											{ return Base::begin(); }
		
		iterator end()																{ return Base::end(); }
		const_iterator end() const												{ return Base::end(); }
		
		iterator IteratorTo( Link& p )										{ return Base::IteratorTo( p ); }
		const_iterator IteratorTo( const Link& p ) const				{ return Base::IteratorTo( p ); }
		
		size_type size() const													{ return Base::size(); }
		static size_type max_size()											{ return Base::max_size(); }
		bool empty() const														{ return Base::empty(); }
		
		iterator insert( iterator where, Link& what )					{ return Base::insert( where, what ); }

		template < class InputIterator >
		void insert( iterator w, InputIterator b, InputIterator e )	{ while ( b != e ) insert( w, *b++ ); }

		void splice( iterator where, iterator b, iterator e )			{ Base::splice( where, b, e ); }
		void splice( iterator where, iterator what )						{ Base::splice( where, what ); }

		void splice( iterator where, LinkedList& source )						{ Base::splice( where, source ); }
		void splice( iterator where, LinkedList& s, iterator what )			{ Base::splice( where, s, what ); }
		void splice( iterator w, LinkedList& s, iterator b, iterator e )	{ Base::splice( w, s, b, e ); }

		iterator erase( Link& what )											{ return Base::erase( what ); }
		iterator erase( iterator where )										{ return Base::erase( where ); }
		iterator erase( iterator b, iterator e )							{ return Base::erase( b, e ); }
		
		Link& extract( Link& what )											{ erase( what ); return what; }
		Link& extract( iterator where )										{ return extract( where.Link() ); }
		LinkedList extract( iterator b, iterator e )						{ return LinkedList( *this, b, e ); }
		
		void clear()																{ Base::clear(); }

		void reverse()																{ Base::reverse(); }

		bool operator==( const LinkedList& b ) const
		  {
			const_iterator left = begin();
			const_iterator right = b.begin();
			
			for ( ; left != end() && right != b.end(); ++left, ++right )
				if ( !( *left == *right ) )
					return false;
			
			return left == end() && right == b.end();
		  }

		// Boilerplate comparisons

			bool operator<( const LinkedList& b ) const	{ return std::lexicographical_compare( begin(), end(), b.begin(), b.end() ); }

			bool operator> ( const LinkedList& b ) const					{ return b < *this; }
			bool operator!=( const LinkedList& b ) const					{ return !( *this == b ); }
			bool operator<=( const LinkedList& b ) const					{ return !( *this > b ); }
			bool operator>=( const LinkedList& b ) const					{ return !( *this < b ); }
		
		// Boilerplate optional sequence operations (partial)
			reference front()														{ return *begin(); }
			const_reference front() const										{ return *begin(); }
			void push_front( Link& what )										{ insert( begin(), what ); }
			void pop_front()														{ erase( begin() ); }

			reference back()														{ return *--end(); }
			const_reference back() const										{ return *--end(); }
			void push_back( Link& what )										{ insert( end(), what ); }
			void pop_back()														{ erase( --end() ); }
		
		// Boilerplate reversibility
			typedef std::reverse_iterator< iterator >			reverse_iterator;
			typedef std::reverse_iterator< const_iterator > const_reverse_iterator;

			reverse_iterator rbegin()											{ return reverse_iterator( end() ); }
			const_reverse_iterator rbegin() const							{ return const_reverse_iterator( end() ); }
			
			reverse_iterator rend()												{ return reverse_iterator( begin() ); }
			const_reverse_iterator rend() const								{ return const_reverse_iterator( begin() ); }
		

		LinkedList remove( const_reference r )								{ return remove_if( std::bind2nd( std::equal_to<value_type>(), r ) ); }
		LinkedList unique()														{ return unique( std::equal_to<value_type>() ); }
		void merge( LinkedList& b )											{ merge( b, std::less<value_type>() ); }
		void sort()																	{ sort( std::less<value_type>() ); }
		
		template < class Unary > LinkedList remove_if( Unary test )
		  {
			LinkedList removed;
			iterator scanned = begin();

			while ( scanned != end() )
			  {
				iterator startRemoving = std::find_if( scanned,       end(), test   );
				iterator stopRemoving  = std::find_if( startRemoving, end(), std::not1( test ) );
				removed.splice( removed.end(), *this, startRemoving, stopRemoving );
				scanned = stopRemoving;
			  }
			
			Assert( std::find_if( begin(), end(), test ) == end() );
			Assert( std::find_if( removed.begin(), removed.end(), std::not1( test ) ) == removed.end() );
			
			return removed;
		  }
		
		template < class Binary > LinkedList unique( Binary test )
		  {
			LinkedList removed;
			iterator scanned = begin();

			while ( scanned != end() )
			  {
				iterator startRemoving = std::adjacent_find( scanned, end(), test );
				iterator stopRemoving  = std::find_if( startRemoving, end(), std::not1( std::bind1st( test, *startRemoving ) ) );
					// There's some debate going on about whet to do when test is not
					// an equivalence relation.  The other likely algorithm is
					// stopRemoving = std::adjacent_find( startRemoving, end(), std::not2( test ) )
				removed.splice( removed.end(), *this, startRemoving, stopRemoving );
				scanned = stopRemoving;
			  }

			Assert( std::adjacent_find( begin(), end(), test ) == end() );
				// If the debate goes the other way, this is not the correct postcondition.
			
			return removed;
		  }
		
		template < class Comparison > void merge( LinkedList& b, Comparison less )
		  {
			iterator scanned = begin();
			
			while ( !b.empty() )
			  {
				scanned = std::find_if( scanned, end(), std::bind1st( less, b.front() ) );

				iterator firstNotLess = ( scanned == end() )
											   ? b.end()
											   : std::find_if( b.begin(), b.end(),
																	 std::not1( std::bind2nd( less, *scanned ) ) );
				
				splice( scanned, b, b.begin(), firstNotLess );
			  }
		  }

		template < class Comparison > void sort( Comparison less )
		  {
			typedef Array< LinkedList, numeric_limits<size_type>::digits > ListArray;
				// We should only need that many because, when full,
				// slot n will contain at least 2^n items.
			ListArray sortedPieces;
			ListArray::iterator usedPiecesEnd = sortedPieces.begin();
			
			while ( true )
			  {
				reverse_iterator lastDrop = std::adjacent_find( rbegin(), rend(), less );

				if ( lastDrop == rend() )
					break;
				
				LinkedList carry( *this, --lastDrop.base(), end() );
				
				ListArray::iterator piece = sortedPieces.begin();
				while ( !piece->empty() )
				  {
					carry.merge( *piece, less );
					++piece;
					Assert( piece != sortedPieces.end() );
				  }
				
				*piece = carry;
				
				Assert( piece <= usedPiecesEnd );
				if ( piece == usedPiecesEnd )
					++usedPiecesEnd;
			  }
			
			for ( ListArray::iterator piece = sortedPieces.begin();
					piece != usedPiecesEnd;
					piece++ )
				merge( *piece, less );

			Assert( std::adjacent_find( rbegin(), rend(), less ) == rend() );
		  }
		
  };

namespace std
  {
	template < class T >
	inline void swap( typename ::LinkedList<T>::Position& a,
							typename ::LinkedList<T>::Position& b )
	  {
		a.swap( b );
	  }

	template < class T >
	inline void swap( typename ::LinkedList<T>::Link& a,
							typename ::LinkedList<T>::Link& b )
	  {
		a.swap( b );
	  }

	template < class T >
	inline void swap( typename ::LinkedList<T>& a,
							typename ::LinkedList<T>& b )
	  {
		a.swap( b );
	  }
  }

#endif
