// LinkedListBase.h

#ifndef LinkedListBase_h
#define LinkedListBase_h

#include <algorithm>
#include <limits>
#include <iterator>
#ifndef ResourceTransfer_h
#include "ResourceTransfer.h"
#endif
#ifndef RingLink_h
#include "RingLink.h"
#endif
#ifndef Assert_h
#include "Assert.h"
#endif

template < bool isSafe > class LinkedListBase;

template <>
class LinkedListBase< false >
  {
	public:
		class Link;
		class const_iterator;
		class iterator;

		typedef std::size_t				size_type;
		typedef std::ptrdiff_t			difference_type;
		typedef Link						value_type;
		typedef Link&						reference;
		typedef const Link&				const_reference;
		
		class Link: private RingLink
		  {
			friend class LinkedListBase;
			friend class const_iterator;
			friend class iterator;

			private:
				Link *Next()														{ return static_cast<       Link * >( RingLink::Next() ); }
				const Link *Next() const										{ return static_cast< const Link * >( RingLink::Next() ); }

				Link *Previous()													{ return static_cast<       Link * >( RingLink::Previous() ); }
				const Link *Previous() const									{ return static_cast< const Link * >( RingLink::Previous() ); }
			
			public:
				Link()																{}
				Link( Link& s )													: RingLink(  s )		{}
				Link( ResourceTransfer<Link> s )								: RingLink( *s )		{}

				operator ResourceTransfer<Link>()							{ return *this; }

				Link& operator=( Link& s )										{ RingLink::operator=( s ); return *this; }
				Link& operator=( ResourceTransfer<Link> s )				{ return operator=( *s ); }

				void swap( Link& s )												{ RingLink::swap( s ); }
				
				using RingLink::Unlink;
		  };
		
		class const_iterator
		  {
			friend class iterator;
			friend class LinkedListBase;
			
			private:
				typedef LinkedListBase::Link LinkType;
				
				const LinkedListBase::Link *p;

				explicit const_iterator( const LinkType *link )			: p( link )			{}

			public:
				typedef std::ptrdiff_t							difference_type;
				typedef LinkType									value_type;
				typedef const LinkType *						pointer;
				typedef const LinkType&							reference;
				typedef std::bidirectional_iterator_tag	iterator_category;
				
				const_iterator()													: p( 0 )				{}
				
				const LinkType& operator*() const							{ return *p; }
				const LinkType *operator->() const							{ return p; }
				
				const_iterator& operator++()									{ p = p->Next();     return *this; }
				const_iterator& operator--()									{ p = p->Previous(); return *this; }

				const_iterator operator++(int)								{ const_iterator old = *this; ++*this; return old; }				
				const_iterator operator--(int)								{ const_iterator old = *this; --*this; return old; }
				
				bool operator==( const const_iterator& b ) const		{ return p == b.p; }
				bool operator!=( const const_iterator& b ) const		{ return p != b.p; }
		  };
		
		class iterator
		  {
			friend class LinkedListBase;
			
			private:
				typedef LinkedListBase::Link LinkType;
				
				LinkedListBase::Link *p;

				explicit iterator( LinkType *link )							: p( link )			{}

			public:
				typedef std::ptrdiff_t							difference_type;
				typedef LinkType									value_type;
				typedef LinkType *								pointer;
				typedef LinkType&									reference;
				typedef std::bidirectional_iterator_tag	iterator_category;

				iterator()															: p( 0 )				{}
				
				operator const_iterator() const								{ return const_iterator( p ); }
				
				LinkType& operator*() const									{ return *p; }
				LinkType *operator->() const									{ return p; }
				
				iterator& operator++()											{ p = p->Next();     return *this; }
				iterator& operator--()											{ p = p->Previous(); return *this; }

				iterator operator++(int)										{ iterator old = *this; ++*this; return old; }				
				iterator operator--(int)										{ iterator old = *this; --*this; return old; }
				
				bool operator==( const const_iterator& b ) const		{ return const_iterator( *this ) == b; }
				bool operator!=( const const_iterator& b ) const		{ return const_iterator( *this ) != b; }
		  };
		
	private:
		Link endLink;
	
	public:
		LinkedListBase()																	{}
		LinkedListBase( LinkedListBase& s )											: endLink( s.endLink )	{}
		LinkedListBase( ResourceTransfer<LinkedListBase> s )					: endLink( s->endLink )	{}
		LinkedListBase( LinkedListBase& s, iterator what );
		LinkedListBase( LinkedListBase& s, iterator b, iterator e );
		
		~LinkedListBase()																	{ clear(); }
		
		LinkedListBase& operator=( LinkedListBase& s )							{ endLink = s.endLink; return *this; }
		LinkedListBase& operator=( ResourceTransfer<LinkedListBase> s )	{ return operator=( *s ); }

		void swap( LinkedListBase& source )											{ endLink.swap( source.endLink ); }

		operator ResourceTransfer<LinkedListBase>()								{ return *this; }

		iterator begin()												   			   { return       iterator( endLink.Next() ); }
		const_iterator begin() const									      		{ return const_iterator( endLink.Next() ); }
		
		iterator end()										      						{ return       iterator( &endLink ); }
		const_iterator end() const								      				{ return const_iterator( &endLink ); }
		
		iterator IteratorTo( Link& p )								      		{ return       iterator( &p ); }
		const_iterator IteratorTo( const Link& p ) const				      { return const_iterator( &p ); }
		
		size_type size() const;
		static size_type max_size()					      						{ return std::numeric_limits<size_type>::max(); }
		bool empty() const										      				{ return begin() == end(); }
		
		iterator insert( iterator where, Link& what );

		void splice( iterator where, iterator b, iterator e );
		void splice( iterator where, iterator what )				      		{ splice( where, what, ++what ); }

		void splice( iterator where, LinkedListBase& source )					{ splice( where, source.begin(), source.end() ); }
		void splice( iterator where, LinkedListBase&, iterator what )		{ splice( where, what ); }
		void splice( iterator w, LinkedListBase&, iterator b, iterator e )	{ splice( w, b, e ); }
		
		iterator erase( Link& what )											      { return erase( IteratorTo( what ) ); }
		iterator erase( iterator where )						      				{ (where++)->Unlink(); return where; }
		iterator erase( iterator b, iterator e );
		
		Link& extract( Link& what )									      		{ erase( what ); return what; }
		Link& extract( iterator where )										      { return extract( *where ); }
		LinkedListBase extract( iterator b, iterator e )						{ return LinkedListBase( *this, b, e ); }
		
		void clear()													      			{ erase( begin(), end() ); }
		
		void reverse();

		// Boilerplate optional sequence operations (partial)
			reference front()													      	{ return *begin(); }
			const_reference front() const				      						{ return *begin(); }
			void push_front( Link& what )						      				{ insert( begin(), what ); }
			void pop_front()												      		{ erase( begin() ); }

			reference back()																{ return *--end(); }
			const_reference back() const												{ return *--end(); }
			void push_back( Link& what )												{ insert( end(), what ); }
			void pop_back()																{ erase( --end() ); }
		
		// Boilerplate reversibility
			typedef std::reverse_iterator< iterator >			reverse_iterator;
			typedef std::reverse_iterator< const_iterator > const_reverse_iterator;

			reverse_iterator rbegin()													{ return reverse_iterator( end() ); }
			const_reverse_iterator rbegin() const									{ return const_reverse_iterator( end() ); }
			
			reverse_iterator rend()														{ return reverse_iterator( begin() ); }
			const_reverse_iterator rend() const										{ return const_reverse_iterator( begin() ); }
  };

template <>
class LinkedListBase< true >
  {
	public:
		class Link;
		class const_iterator;
		class iterator;

		typedef std::size_t				size_type;
		typedef std::ptrdiff_t			difference_type;
		typedef Link						value_type;
		typedef Link&						reference;
		typedef const Link&				const_reference;
		
		class Link: private RingLink
		  {
			friend class LinkedListBase;
			friend class const_iterator;
			friend class iterator;
			
			private:
				const Link *end;
				
				static Link singular;
				
				Link *Next()														{ return static_cast<       Link * >( RingLink::Next() ); }
				const Link *Next() const										{ return static_cast< const Link * >( RingLink::Next() ); }

				Link *Previous()													{ return static_cast<       Link * >( RingLink::Previous() ); }
				const Link *Previous() const									{ return static_cast< const Link * >( RingLink::Previous() ); }
			
			public:
				Link();
				Link( Link& s );
				Link( ResourceTransfer<Link> s );

				operator ResourceTransfer<Link>()							{ return *this; }

				Link& operator=( Link& s );
				Link& operator=( ResourceTransfer<Link> s )				{ return operator=( *s ); }

				void swap( Link& s );
				
				void Unlink();
		  };
		
		class const_iterator
		  {
			friend class iterator;
			friend class LinkedListBase;
			
			private:
				typedef LinkedListBase::Link LinkType;
				
				const LinkedListBase::Link *p;

				explicit const_iterator( const LinkType *link )			: p( link )						{}

				bool HasNext() const												{ return p != p->end; }
				bool HasPrevious() const										{ return p->Previous() != p->end; }
				bool HasStar() const												{ return p != p->end; }
				bool Singular() const											{ return p == &Link::singular; }
				bool ComparableTo( const const_iterator& b ) const		{ return p->end == b.p->end && !Singular(); }
				
			public:
				typedef std::ptrdiff_t							difference_type;
				typedef LinkType									value_type;
				typedef const LinkType *						pointer;
				typedef const LinkType&							reference;
				typedef std::bidirectional_iterator_tag	iterator_category;

				const_iterator()													: p( &Link::singular )		{}
				
				const LinkType& operator*() const							{ Assert( HasStar() ); return *p; }
				const LinkType *operator->() const							{ Assert( HasStar() ); return p; }
				
				const_iterator& operator++()									{ Assert( HasNext() );     p = p->Next();     return *this; }
				const_iterator& operator--()									{ Assert( HasPrevious() ); p = p->Previous(); return *this; }

				const_iterator operator++(int)								{ const_iterator old = *this; ++*this; return old; }				
				const_iterator operator--(int)								{ const_iterator old = *this; --*this; return old; }
				
				bool operator==( const const_iterator& b ) const		{ Assert( ComparableTo(b) ); return p == b.p; }
				bool operator!=( const const_iterator& b ) const		{ Assert( ComparableTo(b) ); return p != b.p; }
		  };
		
		class iterator
		  {
			friend class LinkedListBase;
			
			private:
				typedef LinkedListBase::Link LinkType;
				
				LinkedListBase::Link *p;

				explicit iterator( LinkType *link )							: p( link )			{}

				bool HasNext() const												{ return p != p->end; }
				bool HasPrevious() const										{ return p->Previous() != p->end; }
				bool HasStar() const												{ return p != p->end; }
				bool Singular() const											{ return p == &Link::singular; }
				bool ComparableTo( const const_iterator& b ) const		{ return p->end == b.p->end && !Singular(); }

			public:
				typedef std::ptrdiff_t							difference_type;
				typedef LinkType									value_type;
				typedef LinkType *								pointer;
				typedef LinkType&									reference;
				typedef std::bidirectional_iterator_tag	iterator_category;

				iterator()															: p( &Link::singular )				{}
				
				operator const_iterator() const								{ return const_iterator( p ); }
				
				LinkType& operator*() const									{ Assert( HasStar() ); return *p; }
				LinkType *operator->() const									{ Assert( HasStar() ); return p; }
				
				iterator& operator++()											{ Assert( HasNext() );     p = p->Next();     return *this; }
				iterator& operator--()											{ Assert( HasPrevious() ); p = p->Previous(); return *this; }

				iterator operator++(int)										{ iterator old = *this; ++*this; return old; }				
				iterator operator--(int)										{ iterator old = *this; --*this; return old; }
				
				bool operator==( const const_iterator& b ) const		{ return const_iterator( *this ) == b; }
				bool operator!=( const const_iterator& b ) const		{ return const_iterator( *this ) != b; }
		  };
		
	private:
		Link endLink;
	
		bool Contains( const Link& link ) const							{ return link.end == &endLink; }
		bool Contains( const_iterator p ) const							{ return p.ComparableTo( end() ); }
		static bool IsValidRange( const_iterator b, const_iterator e );
		void FixEnds();
		
	public:
		LinkedListBase()																	{}
		LinkedListBase( LinkedListBase& s );
		LinkedListBase( ResourceTransfer<LinkedListBase> s );
		LinkedListBase( LinkedListBase& s, iterator what );
		LinkedListBase( LinkedListBase& s, iterator b, iterator e );
		
		template < class InputIterator >
		LinkedListBase( InputIterator b, InputIterator e )						{ while ( b != e ) push_back( *b++ ); }
		
		~LinkedListBase()																	{ clear(); }
		
		LinkedListBase& operator=( LinkedListBase& s );
		LinkedListBase& operator=( ResourceTransfer<LinkedListBase> s )	{ return operator=( *s ); }

		void swap( LinkedListBase& source );

		operator ResourceTransfer<LinkedListBase>()								{ return *this; }

		iterator begin()																	{ return       iterator( endLink.Next() ); }
		const_iterator begin() const													{ return const_iterator( endLink.Next() ); }
		
		iterator end()																		{ return       iterator( &endLink ); }
		const_iterator end() const														{ return const_iterator( &endLink ); }
		
		iterator IteratorTo( Link& p )												{ Assert( Contains( p ) ); return       iterator( &p ); }
		const_iterator IteratorTo( const Link& p ) const						{ Assert( Contains( p ) ); return const_iterator( &p ); }
		
		size_type size() const;
		static size_type max_size()													{ return std::numeric_limits<size_type>::max(); }
		bool empty() const																{ return begin() == end(); }

		iterator insert( iterator where, Link& what );

		template < class InputIterator >
		void insert( iterator w, InputIterator b, InputIterator e )			{ while ( b != e ) insert( w, *b++ ); }

		void splice( iterator where, iterator b, iterator e );
		void splice( iterator where, iterator what )								{ splice( where, what, ++what ); }

		void splice( iterator where, LinkedListBase& source )					{ splice( where, source.begin(), source.end() ); }
		void splice( iterator where, LinkedListBase&, iterator what );
		void splice( iterator w, LinkedListBase&, iterator b, iterator e );

		iterator erase( Link& what )													{ return erase( IteratorTo( what ) ); }
		iterator erase( iterator where );
		iterator erase( iterator b, iterator e );
		
		Link& extract( Link& what )													{ erase( what ); return what; }
		Link& extract( iterator where )												{ return extract( *where ); }
		LinkedListBase extract( iterator b, iterator e )						{ return LinkedListBase( *this, b, e ); }
		
		void clear()																		{ erase( begin(), end() ); }
		
		void reverse();

		// Boilerplate optional sequence operations (partial)
			reference front()																{ return *begin(); }
			const_reference front() const												{ return *begin(); }
			void push_front( Link& what )												{ insert( begin(), what ); }
			void pop_front()																{ erase( begin() ); }

			reference back()																{ return *--end(); }
			const_reference back() const												{ return *--end(); }
			void push_back( Link& what )												{ insert( end(), what ); }
			void pop_back()																{ erase( --end() ); }
		
		// Boilerplate reversibility
			typedef std::reverse_iterator< iterator >			reverse_iterator;
			typedef std::reverse_iterator< const_iterator > const_reverse_iterator;

			reverse_iterator rbegin()													{ return reverse_iterator( end() ); }
			const_reverse_iterator rbegin() const									{ return const_reverse_iterator( end() ); }
			
			reverse_iterator rend()														{ return reverse_iterator( begin() ); }
			const_reverse_iterator rend() const										{ return const_reverse_iterator( begin() ); }
  };

namespace std
  {
	template <>
	inline void swap( ::LinkedListBase<false>::Link& a, ::LinkedListBase<false>::Link& b )
	  {
		a.swap( b );
	  }

	template <>
	inline void swap( ::LinkedListBase<false>& a, ::LinkedListBase<false>& b )
	  {
		a.swap( b );
	  }

	template <>
	inline void swap( ::LinkedListBase<true>::Link& a, ::LinkedListBase<true>::Link& b )
	  {
		a.swap( b );
	  }

	template <>
	inline void swap( ::LinkedListBase<true>& a, ::LinkedListBase<true>& b )
	  {
		a.swap( b );
	  }
  }

#endif
