// Observable.h

#ifndef Observable_h
#define Observable_h

#ifndef Closure_h
#include "Closure.h"
#endif
#ifndef LinkedList_h
#include "LinkedList.h"
#endif
#ifndef Observation_h
#include "Observation.h"
#endif

template < class Signature > class Observable;
template < class Signature > class ObservableBase;

template < class Signature >
class ObservableBase
  {
	public:
		typedef ::Closure< Signature > Closure;
		typedef ::Observation< Signature > Observation;
	
	private:
		typedef ::LinkedList< Closure > List;

		mutable List past;
		mutable List future;
		
		// not implemented:
			ObservableBase( const ObservableBase& );
			ObservableBase& operator=( const ObservableBase& );
	
	public:
		ObservableBase()								{}
		
		void StartBroadcast() const				{ future.splice( future.begin(), past ); }
		bool BroadcastDone() const					{ return future.empty(); }
		
		const Closure& NextObserver() const
		  {
			past.splice( past.end(), future, future.begin() );
			return past.back();
		  }

		Observation Notify( const Closure& c ) const
		  {
			Observation observation( c );
			past.insert( past.end(), observation.link );
			return observation;
		  }

		template < class Function >
		Observation Notify( Function function ) const
		  {
			Observation observation = Closure( function );
			past.insert( past.end(), observation.link );
			return observation;
		  }
		
		template < class Object, class Method >
		Observation Notify( Object& object, Method method ) const
		  {
			Observation observation( Closure( object, method ) );
			past.insert( past.end(), observation.link );
			return observation;
		  }
		
		template < class Object, class Method >
		Observation Notify( const Object& object, Method method ) const
		  {
			Observation observation( Closure( object, method ) );
			past.insert( past.end(), observation.link );
			return observation;
		  }
  };


template <>
class Observable< void (*)() >: private ObservableBase< void(*)() >
  {
	private:
		typedef ObservableBase< void(*)() > Base;
		
	public:
		typedef typename Base::Closure Closure;
		typedef typename Base::Observation Observation;
		
	protected:
		Observable()				{}
		~Observable()				{}
		
		using Base::StartBroadcast;
		using Base::BroadcastDone;
		using Base::NextObserver;
		
		void ContinueBroadcast() const
		  {
			while ( !BroadcastDone() )
				NextObserver()();
		  }
		
		void Broadcast() const
		  {
			StartBroadcast();
			ContinueBroadcast();
		  }
	
	public:
		using Base::Notify;
  };


template < class P0 >
class Observable< void (*)( P0 ) >: private ObservableBase< void(*)( P0 ) >
  {
	private:
		typedef ObservableBase< void(*)( P0 ) > Base;

	public:
		typedef typename Base::Closure Closure;
		typedef typename Base::Observation Observation;
		
	protected:
		Observable()				{}
		~Observable()				{}
		
		using Base::StartBroadcast;
		using Base::BroadcastDone;
		using Base::NextObserver;
		
		void ContinueBroadcast( P0 p0 ) const
		  {
			while ( !BroadcastDone() )
				NextObserver()( p0 );
		  }
		
		void Broadcast( P0 p0 ) const
		  {
			StartBroadcast();
			ContinueBroadcast( p0 );
		  }
	
	public:
		using Base::Notify;
  };


template < class P0, class P1 >
class Observable< void (*)( P0, P1 ) >: private ObservableBase< void(*)( P0, P1 ) >
  {
	private:
		typedef ObservableBase< void(*)( P0, P1 ) > Base;
	
	public:
		typedef typename Base::Closure Closure;
		typedef typename Base::Observation Observation;
		
	protected:
		Observable()				{}
		~Observable()				{}
		
		using Base::StartBroadcast;
		using Base::BroadcastDone;
		using Base::NextObserver;
		
		void ContinueBroadcast( P0 p0, P1 p1 ) const
		  {
			while ( !BroadcastDone() )
				NextObserver()( p0, p1 );
		  }
		
		void Broadcast( P0 p0, P1 p1 ) const
		  {
			StartBroadcast();
			ContinueBroadcast( p0, p1 );
		  }
	
	public:
		using Base::Notify;
  };


template < class P0, class P1, class P2 >
class Observable< void (*)( P0, P1, P2 ) >: private ObservableBase< void(*)( P0, P1, P2 ) >
  {
	private:
		typedef ObservableBase< void(*)( P0, P1, P2 ) > Base;

	public:
		typedef typename Base::Closure Closure;
		typedef typename Base::Observation Observation;
		
	protected:
		Observable()				{}
		~Observable()				{}
		
		using Base::StartBroadcast;
		using Base::BroadcastDone;
		using Base::NextObserver;
		
		void ContinueBroadcast( P0 p0, P1 p1, P2 p2 ) const
		  {
			while ( !BroadcastDone() )
				NextObserver()( p0, p1, p2 );
		  }
		
		void Broadcast( P0 p0, P1 p1, P2 p2 ) const
		  {
			StartBroadcast();
			ContinueBroadcast( p0, p1, p2 );
		  }
	
	public:
		using Base::Notify;
  };


template < class P0, class P1, class P2, class P3 >
class Observable< void (*)( P0, P1, P2, P3 ) >: private ObservableBase< void(*)( P0, P1, P2, P3 ) >
  {
	private:
		typedef ObservableBase< void(*)( P0, P1, P2, P3 ) > Base;

	public:
		typedef typename Base::Closure Closure;
		typedef typename Base::Observation Observation;
		
	protected:
		Observable()				{}
		~Observable()				{}
		
		using Base::StartBroadcast;
		using Base::BroadcastDone;
		using Base::NextObserver;
		
		void ContinueBroadcast( P0 p0, P1 p1, P2 p2, P3 p3 ) const
		  {
			while ( !BroadcastDone() )
				NextObserver()( p0, p1, p2, p3 );
		  }
		
		void Broadcast( P0 p0, P1 p1, P2 p2, P3 p3 ) const
		  {
			StartBroadcast();
			ContinueBroadcast( p0, p1, p2, p3 );
		  }
	
	public:
		using Base::Notify;
  };

#endif
